@@ -5,14 +5,14 @@
 
 #define _GNU_SOURCE
 #include <argp.h>
-#include <fnmatch.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <unistd.h>
 
 #include "config.h"
-#include "ch_misc.h"
 #include "ch_core.h"
+#include "ch_misc.h"
 
 
 /** Constants and macros **/
@@ -39,7 +39,7 @@ Example:\n\
 \n\
 You cannot use this program to actually change your UID.\n";
 
-const char args_doc[] = "NEWROOT CMD [ARG...]";
+const char args_doc[] = "IMAGE -- CMD [ARG...]";
 
 const struct argp_option options[] = {
    { "bind",          'b', "SRC[:DST]", 0,
@@ -52,10 +52,12 @@ const struct argp_option options[] = {
    { "join-pid",       -5, "PID",  0, "join a namespace using a PID" },
    { "join-ct",        -3, "N",    0, "number of join peers (implies --join)" },
    { "join-tag",       -4, "TAG",  0, "label for peer group (implies --join)" },
+   { "mount",         'm', "DIR",  0, "SquashFS mount point"},
    { "no-home",        -2, 0,      0, "don't bind-mount your home directory"},
    { "no-passwd",      -9, 0,      0, "don't bind-mount /etc/{passwd,group}"},
    { "private-tmp",   't', 0,      0, "use container-private /tmp" },
-   { "set-env",        -6, "FILE", 0, "set environment variables in FILE"},
+   { "set-env",        -6, "ARG",  OPTION_ARG_OPTIONAL,
+     "set environment variables per ARG"},
    { "uid",           'u', "UID",  0, "run as UID within container" },
    { "unset-env",      -7, "GLOB", 0, "unset environment variable(s)" },
    { "verbose",       'v', 0,      0, "be more verbose (debug if repeated)" },
@@ -64,12 +66,8 @@ const struct argp_option options[] = {
    { 0 }
 };
 
-enum env_action { ENV_END = 0, ENV_SET, ENV_UNSET };
 
-struct env_delta {
-   enum env_action action;
-   char *arg;
-};
+/** Types **/
 
 struct args {
    struct container c;
@@ -77,11 +75,9 @@ struct args {
    char *initial_dir;
 };
 
+
 /** Function prototypes **/
 
-void env_delta_append(struct env_delta **ds, enum env_action act, char *arg);
-void envs_set(char **lines, const int line_ct, const char *filename,
-              const bool expand);
 void fix_environment(struct args *args);
 bool get_first_env(char **array, char **name, char **value);
 int join_ct(int cli_ct);
@@ -104,30 +100,36 @@ int main(int argc, char *argv[])
    bool argp_help_fmt_set;
    struct args args;
    int arg_next;
-   int c_argc;
    char ** c_argv;
 
    privs_verify_invoking();
 
+#ifdef ENABLE_SYSLOG
+   syslog(LOG_USER|LOG_INFO, "uid=%u args=%d: %s", getuid(), argc,
+          argv_to_string(argv));
+#endif
+
    verbose = 1;  // in charliecloud.h
-   args = (struct args){ .c = (struct container){ .ch_ssh = false,
-                                                  .container_gid = getegid(),
-                                                  .container_uid = geteuid(),
-                                                  .env_expand = true,
-                                                  .newroot = NULL,
-                                                  .join = false,
-                                                  .join_ct = 0,
-                                                  .join_pid = 0,
-                                                  .join_tag = NULL,
-                                                  .private_home = false,
-                                                  .private_passwd = false,
-                                                  .private_tmp = false,
-                                                  .old_home = getenv("HOME"),
-                                                  .writable = false },
-                         .initial_dir = NULL };
-   // These need to be on the heap because we realloc(3) them later.
-   T_ (args.c.binds = calloc(1, sizeof(struct bind)));
-   T_ (args.env_deltas = calloc(1, sizeof(struct env_delta)));
+   args = (struct args){
+      .c = (struct container){ .binds = list_new(sizeof(struct bind), 0),
+                               .ch_ssh = false,
+                               .container_gid = getegid(),
+                               .container_uid = geteuid(),
+                               .env_expand = true,
+                               .img_path = NULL,
+                               .newroot = NULL,
+                               .join = false,
+                               .join_ct = 0,
+                               .join_pid = 0,
+                               .join_tag = NULL,
+                               .private_home = false,
+                               .private_passwd = false,
+                               .private_tmp = false,
+                               .old_home = getenv("HOME"),
+                               .type = IMG_NONE,
+                               .writable = false },
+      .env_deltas = list_new(sizeof(struct env_delta), 0),
+      .initial_dir = NULL };
 
    /* I couldn't find a way to set argp help defaults other than this
       environment variable. Kludge sets/unsets only if not already set. */
@@ -142,21 +144,44 @@ int main(int argc, char *argv[])
       Z_ (unsetenv("ARGP_HELP_FMT"));
 
    Te (arg_next < argc - 1, "NEWROOT and/or CMD not specified");
-   args.c.newroot = realpath(argv[arg_next], NULL);
-   Tf (args.c.newroot != NULL, "can't find image: %s", argv[arg_next]);
-   arg_next++;
+   args.c.img_path = argv[arg_next++];
+   args.c.type = img_type_get(args.c.img_path);
+
+   switch (args.c.type) {
+   case IMG_DIRECTORY:
+      if (args.c.newroot != NULL)  // --mount was set
+         WARNING("--mount invalid with directory image, ignoring");
+      args.c.newroot = realpath(args.c.img_path, NULL);
+      Tf (args.c.newroot != NULL, "can't find image: %s", args.c.img_path);
+      break;
+   case IMG_SQUASH:
+#ifndef HAVE_LIBSQUASHFUSE
+      FATAL("this ch-run does not support internal SquashFS mounts");
+#endif
+      break;
+   case IMG_NONE:
+      FATAL("unknown image type: %s", args.c.img_path);
+      break;
+   }
 
    if (args.c.join) {
       args.c.join_ct = join_ct(args.c.join_ct);
       args.c.join_tag = join_tag(args.c.join_tag);
    }
 
-   c_argc = argc - arg_next;
-   T_ (c_argv = calloc(c_argc + 1, sizeof(char *)));
-   for (int i = 0; i < c_argc; i++)
+   if (getenv("TMPDIR") != NULL)
+      host_tmp = getenv("TMPDIR");
+   else
+      host_tmp = "/tmp";
+   username = getenv("USER");
+   Te (username != NULL, "$USER not set");
+
+   c_argv = list_new(sizeof(char *), argc - arg_next);
+   for (int i = 0; i < argc - arg_next; i++)
       c_argv[i] = argv[i + arg_next];
 
    INFO("verbosity: %d", verbose);
+   INFO("image: %s", args.c.img_path);
    INFO("newroot: %s", args.c.newroot);
    INFO("container uid: %u", args.c.container_uid);
    INFO("container gid: %u", args.c.container_gid);
@@ -164,8 +189,8 @@ int main(int argc, char *argv[])
         args.c.join_pid);
    INFO("private /tmp: %d", args.c.private_tmp);
 
-   fix_environment(&args);
    containerize(&args.c);
+   fix_environment(&args);
    run_user_command(c_argv, args.initial_dir); // should never return
    exit(EXIT_FAILURE);
 }
@@ -173,95 +198,15 @@ int main(int argc, char *argv[])
 
 /** Supporting functions **/
 
-/* Append a new env_delta to an existing null-terminated list. */
-void env_delta_append(struct env_delta **ds, enum env_action act, char *arg)
-{
-   int i;
-
-   for (i = 0; (*ds)[i].action != ENV_END; i++) // count existing
-         ;
-   T_ (*ds = realloc(*ds, (i+2) * sizeof(struct env_delta)));
-   (*ds)[i+1].action = ENV_END;
-   (*ds)[i].action = act;
-   (*ds)[i].arg = arg;
-}
-
-/* Set environment variables as specified in the array lines, which has length
-   line_ct. filename is the source filename, or NULL if source was not a file;
-   in that case, omit line number from any error messages. If expand, then
-   expand variable notation as described in the man page. */
-void envs_set(char **lines, const int line_ct, const char *filename,
-              const bool expand)
-{
-   char *name, *value_old, *value_new, *lineno_str, *item;
-
-   for (int i = 0; i < line_ct; i++) {
-      bool first_written;
-
-      // Skip blank lines.
-      if (lines[i][0] == 0 || lines[i][0] == '\n')
-         continue;
-
-      // Split line into variable name and value.
-      split(&name, &value_old, lines[i], '=');
-      if (filename == NULL)
-         lineno_str = "";
-      else
-         T_ (1 <= asprintf(&lineno_str, ":%d", i+1));
-      Te (name != NULL, "--set-env: no delimiter: %s%s", filename, lineno_str);
-      Te (name[0] != 0, "--set-env: empty name: %s%s", filename, lineno_str);
-
-      // Strip leading and trailing single quotes from value, if both present.
-      if (   strlen(value_old) >= 2
-          && value_old[0] == '\''
-          && value_old[strlen(value_old) - 1] == '\'') {
-         value_old[strlen(value_old) - 1] = 0;
-         value_old++;
-      }
-
-      // Walk through value fragments separated by colon and expand variables
-      // per documentation.
-      value_new = "";
-      first_written = false;
-      while (1) {                                  // loop executes ≥ once
-         item = strsep(&value_old, ":");           // NULL -> no more items
-         if (item == NULL)
-            break;
-         if (   expand                             // expansion requested
-             && item[0] == '$' && item[1] != 0) {  // ≥1 char in variable name
-            item = getenv(++item);                 // NULL if unset
-            if (item != NULL && item[0] == 0)
-               item = NULL;                        // convert empty to unset
-         }
-         if (item != NULL) {                       // NULL -> omit from output
-            if (first_written)
-               value_new = cat(value_new, ":");
-            value_new = cat(value_new, item);
-            first_written = true;
-         }
-      }
-
-      // Save results.
-      INFO("environment: %s=%s", name, value_new);
-      Z_ (setenv(name, value_new, 1));
-   }
-}
-
-/* Adjust environment variables. */
+/* Adjust environment variables. Call once containerized, i.e., already
+   pivoted into new root. */
 void fix_environment(struct args *args)
 {
-   char *name, *old_value, *new_value;
+   char *old_value, *new_value;
 
    // $HOME: Set to /home/$USER unless --no-home specified.
-   if (!args->c.private_home) {
-      old_value = getenv("USER");
-      if (old_value == NULL) {
-         WARNING("$USER not set; cannot rewrite $HOME");
-      } else {
-         T_ (1 <= asprintf(&new_value, "/home/%s", old_value));
-         Z_ (setenv("HOME", new_value, 1));
-      }
-   }
+   if (!args->c.private_home)
+      Z_ (setenv("HOME", cat("/home/", username), 1));
 
    // $PATH: Append /bin if not already present.
    old_value = getenv("PATH");
@@ -274,73 +219,31 @@ void fix_environment(struct args *args)
       INFO("new $PATH: %s", new_value);
    }
 
+   // $TMPDIR: Unset.
+   Z_ (unsetenv("TMPDIR"));
+
    // --set-env and --unset-env.
-   for (int i = 0; args->env_deltas[i].action != ENV_END; i++) {
-      char *arg = args->env_deltas[i].arg;
-      if (args->env_deltas[i].action == ENV_SET) {  // --set-env
-         if (strchr(arg, '=') != NULL) {
-            // argument is variable name & value
-            envs_set((char *[]){ arg }, 1, NULL, args->c.env_expand);
-         } else {
-            // argument is filename
-            char **lines = NULL;
-            int line_ct = 0;
-            FILE *fp;
-            Tf (fp = fopen(arg, "r"), "--set-env: can't open: %s", arg);
-            for (line_ct = 0; true; line_ct++) {
-               char *line;
-               size_t line_len = 0;  // don't care but getline(3) must write
-               errno = 0;
-               if (-1 == getline(&line, &line_len, fp)) {
-                  if (errno == 0)  // EOF
-                     break;        // note: line_ct not incremented
-                  else
-                     Tf (0, "--set-env: can't read: %s", arg);
-               }
-               if (line[strlen(line) - 1] == '\n')  // rm newline if present
-                  line[strlen(line) - 1] = 0;
-               T_ (lines = realloc(lines, (line_ct + 1) * sizeof(char *)));
-               lines[line_ct] = line;
-            }
-            Zf (fclose(fp), "--set-env: can't close: %s", arg);
-            envs_set(lines, line_ct, arg, args->c.env_expand);
-         }
-      } else {  // --unset-env
-         T_ (args->env_deltas[i].action == ENV_UNSET);
-         /* Removing variables from the environment is tricky, because there
-            is no standard library function to iterate through the
-            environment, and the environ global array can be re-ordered after
-            unsetenv(3) [1]. Thus, the only safe way without additional
-            storage is an O(n^2) search until no matches remain.
-
-            It is legal to assign to environ [2]. We build up a copy, omitting
-            variables that match the glob, which is O(n), and then do so.
-
-            [1]: https://unix.stackexchange.com/a/302987
-            [2]: http://man7.org/linux/man-pages/man3/exec.3p.html */
-         char **new_environ;
-         int old_i, new_i;
-         for (old_i = 0; environ[old_i] != NULL; old_i++)
-            ;
-         T_ (new_environ = calloc(old_i + 1, sizeof(char *)));
-         for (old_i = 0, new_i = 0; environ[old_i] != NULL; old_i++) {
-            int matchp;
-            split(&name, &old_value, environ[old_i], '=');
-            T_ (name != NULL);          // env lines should always have equals
-            matchp = fnmatch(arg, name, 0);
-            if (!matchp) {
-               INFO("environment: unset %s", name);
-            } else {
-               T_ (matchp == FNM_NOMATCH);
-               *(old_value - 1) = '=';  // rejoin line
-               new_environ[new_i++] = name;
-            }
-         }
-         environ = new_environ;
+   for (size_t i = 0; args->env_deltas[i].action != ENV_END; i++) {
+      struct env_delta ed = args->env_deltas[i];
+      switch (ed.action) {
+      case ENV_END:
+         Te (false, "unreachable code reached");
+         break;
+      case ENV_SET_DEFAULT:
+         ed.arg.vars = env_file_read("/ch/environment");
+         // fall through
+      case ENV_SET_VARS:
+         for (size_t j = 0; ed.arg.vars[j].name != NULL; j++)
+            env_set(ed.arg.vars[j].name, ed.arg.vars[j].value,
+                    args->c.env_expand );
+         break;
+      case ENV_UNSET_GLOB:
+         env_unset(ed.arg.glob);
+         break;
       }
    }
 
-   // $CH_RUNNING
+   // $CH_RUNNING is not affected by --unset-env or --set-env.
    Z_ (setenv("CH_RUNNING", "Weird Al Yankovic", 1));
 }
 
@@ -432,6 +335,7 @@ int parse_int(char *s, bool extra_ok, char *error_tag)
 static error_t parse_opt(int key, char *arg, struct argp_state *state)
 {
    struct args *args = state->input;
+   struct env_delta ed;
    int i;
 
    switch (key) {
@@ -453,11 +357,24 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
       args->c.join_pid = parse_int(arg, false, "--join-pid");
       break;
    case -6: // --set-env
-      env_delta_append(&(args->env_deltas), ENV_SET, arg);
+      if (arg == NULL)
+         ed.action = ENV_SET_DEFAULT;
+      else {
+         ed.action = ENV_SET_VARS;
+         if (strchr(arg, '=') == NULL)
+            ed.arg.vars = env_file_read(arg);
+         else {
+            ed.arg.vars = list_new(sizeof(struct env_var), 1);
+            ed.arg.vars[0] = env_var_parse(arg, NULL, 0);
+         }
+      }
+      list_append((void **)&(args->env_deltas), &ed, sizeof(ed));
       break;
    case -7: // --unset-env
       Te (strlen(arg) > 0, "--unset-env: GLOB must have non-zero length");
-      env_delta_append(&(args->env_deltas), ENV_UNSET, arg);
+      ed.action = ENV_UNSET_GLOB;
+      ed.arg.glob = arg;
+      list_append((void **)&(args->env_deltas), &ed, sizeof(ed));
       break;;
    case -8: // --ch-ssh
       args->c.ch_ssh = true;
@@ -497,6 +414,10 @@ static error_t parse_opt(int key, char *arg, struct argp_state *state)
    case 'j':
       args->c.join = true;
       break;
+   case 'm':
+      Ze ((arg[0] == '\0'), "mount point can't be empty string");
+      args->c.newroot = arg;
+      break;
    case 't':
       args->c.private_tmp = true;
       break;
