@@ -21,6 +21,9 @@
 #include "config.h"
 #include "ch_misc.h"
 #include "ch_core.h"
+#ifdef HAVE_LIBSQUASHFUSE
+#include "ch_fuse.h"
+#endif
 
 
 /** Macros **/
@@ -45,7 +48,7 @@ struct bind BINDS_DEFAULT[] = {
    { "/etc/resolv.conf",         "/etc/resolv.conf",         BD_OPTIONAL },
    { "/var/lib/hugetlbfs",       "/var/opt/cray/hugetlbfs",  BD_OPTIONAL },
    { "/var/opt/cray/alps/spool", "/var/opt/cray/alps/spool", BD_OPTIONAL },
-   { NULL }
+   { 0 }
 };
 
 
@@ -63,9 +66,8 @@ struct {
    } *shared;
 } join;
 
-/* Bind mounts done so far; canonical host paths. */
+/* Bind mounts done so far; canonical host paths. If null, there are none. */
 char **bind_mount_paths = NULL;
-size_t bind_mount_path_ct = 0;
 
 
 /** Function prototypes (private) **/
@@ -75,10 +77,10 @@ void bind_mount(const char *src, const char *dst, enum bind_dep,
 void bind_mounts(const struct bind *binds, const char *newroot,
                  unsigned long flags);
 void enter_udss(struct container *c);
-void join_begin(int join_ct, const char *join_tag);
+void join_begin(const char *join_tag);
 void join_namespace(pid_t pid, const char *ns);
 void join_namespaces(pid_t pid);
-void join_end();
+void join_end(int join_ct);
 void sem_timedwait_relative(sem_t *sem, int timeout);
 void setup_namespaces(const struct container *c);
 void setup_passwd(const struct container *c);
@@ -110,7 +112,7 @@ void bind_mount(const char *src, const char *dst, enum bind_dep dep,
       case BD_OPTIONAL:
          return;
       case BD_MAKE_DST:
-         mkdirs(newroot, dst, bind_mount_paths, bind_mount_path_ct);
+         mkdirs(newroot, dst, bind_mount_paths);
          break;
       }
 
@@ -118,12 +120,8 @@ void bind_mount(const char *src, const char *dst, enum bind_dep dep,
    dst_fullc = realpath_safe(dst_full);
    Tf (path_subdir_p(newrootc, dst_fullc),
        "can't bind: %s not subdirectory of %s", dst_fullc, newrootc);
-   if (strcmp(newroot, "/")) {  // don't record if newroot is "/"
-      bind_mount_path_ct++;
-      T_ (bind_mount_paths = realloc(bind_mount_paths,
-                                     bind_mount_path_ct * sizeof(char *)));
-      bind_mount_paths[bind_mount_path_ct - 1] = dst_fullc;
-   }
+   if (strcmp(newroot, "/"))  // don't record if newroot is "/"
+      list_append((void **)&bind_mount_paths, &dst_fullc, sizeof(char *));
 
    Zf (mount(src, dst_full, NULL, MS_REC|MS_BIND|flags, NULL),
        "can't bind %s to %s", src, dst_full);
@@ -145,14 +143,18 @@ void containerize(struct container *c)
       return;
    }
    if (c->join)
-      join_begin(c->join_ct, c->join_tag);
+      join_begin(c->join_tag);
    if (!c->join || join.winner_p) {
+#ifdef HAVE_LIBSQUASHFUSE
+      if (c->type == IMG_SQUASH)
+         sq_fork(c);
+#endif
       setup_namespaces(c);
       enter_udss(c);
    } else
       join_namespaces(join.shared->winner_pid);
    if (c->join)
-      join_end();
+      join_end(c->join_ct);
 
 }
 
@@ -182,7 +184,7 @@ void enter_udss(struct container *c)
    if (c->private_tmp) {
       tmpfs_mount("/tmp", c->newroot, NULL);
    } else {
-      bind_mount("/tmp", "/tmp", BD_REQUIRED, c->newroot, 0);
+      bind_mount(host_tmp, "/tmp", BD_REQUIRED, c->newroot, 0);
    }
    // Container /home.
    if (!c->private_home) {
@@ -192,10 +194,8 @@ void enter_udss(struct container *c)
       // Bind-mount user's home directory at /home/$USER. The main use case is
       // dotfiles.
       Tf (c->old_home != NULL, "cannot find home directory: is $HOME set?");
-      newhome = cat("/home/", getenv("USER"));
-      if(getenv("USER") != NULL) {
-         Z_ (mkdir(cat(c->newroot, newhome), 0755));
-      }
+      newhome = cat("/home/", username);
+      Z_ (mkdir(cat(c->newroot, newhome), 0755));
       bind_mount(c->old_home, newhome, BD_REQUIRED, c->newroot, 0);
    }
    // Container /usr/bin/ch-ssh.
@@ -235,8 +235,35 @@ void enter_udss(struct container *c)
    Zf (umount2("/dev", MNT_DETACH), "can't umount old root");
 }
 
+/* Return image type of path, or exit with error if not a valid type. */
+enum img_type img_type_get(const char *path)
+{
+   struct stat read;
+   FILE *fp;
+   char magic[4];  // four bytes, not a string
+
+   Zf (stat(path, &read), "can't stat: %s", path);
+
+   if (S_ISDIR(read.st_mode))
+      return IMG_DIRECTORY;
+
+   fp = fopen(path, "rb");
+   Tf (fp != NULL, "can't open: %s", path);
+   Tf (fread(magic, sizeof(char), 4, fp) == 4, "can't read: %s", path);
+   Zf (fclose(fp), "can't close: %s", path);
+   INFO("image file magic expected: 6873 7173; actual: %x%x %x%x", magic[0], magic[1], magic[2], magic[3]);
+
+   // SquashFS magic number is 6873 7173, i.e. "hsqs". I think "sqsh" was
+   // intended but the superblock designers were confused about endianness.
+   // See: https://dr-emann.github.io/squashfs/
+   if (memcmp(magic, "hsqs", 4) == 0)
+      return IMG_SQUASH;
+
+   FATAL("unknown image type: %s", path);
+   return IMG_NONE;  // unreachable, avoid warning; see issue #1158
+}
 /* Begin coordinated section of namespace joining. */
-void join_begin(int join_ct, const char *join_tag)
+void join_begin(const char *join_tag)
 {
    int fd;
 
@@ -255,6 +282,7 @@ void join_begin(int join_ct, const char *join_tag)
       join.winner_p = true;
       Z_ (ftruncate(fd, sizeof(*join.shared)));
    } else if (errno == EEXIST) {
+      INFO("join: I lost");
       join.winner_p = false;
       fd = shm_open(join.shm_name, O_RDWR, 0);
       T_ (fd > 0);
@@ -267,22 +295,19 @@ void join_begin(int join_ct, const char *join_tag)
    T_ (join.shared != NULL);
    Z_ (close(fd));
 
-   if (join.winner_p) {
-      join.shared->winner_pid = getpid();
-      join.shared->proc_left_ct = join_ct;
-      // Keep lock; winner still serialized.
-   } else {
-      INFO("join: winner pid: %d", join.shared->winner_pid);
+   // Winner keeps lock; losers parallelize (winner will be done by now).
+   if (!join.winner_p)
       Z_ (sem_post(join.sem));
-      // Losers run in parallel (winner will be done by now).
-   }
 }
 
 /* End coordinated section of namespace joining. */
-void join_end()
+void join_end(int join_ct)
 {
-   // Serialize (winner never released lock).
-   if (!join.winner_p)
+   if (join.winner_p) {                                // winner still serial
+      INFO("join: winner initializing shared data");
+      join.shared->winner_pid = getpid();
+      join.shared->proc_left_ct = join_ct;
+   } else                                              // losers serialize
       sem_timedwait_relative(join.sem, JOIN_TIMEOUT);
 
    join.shared->proc_left_ct--;
@@ -296,8 +321,7 @@ void join_end()
       Zf (shm_unlink(join.shm_name), "can't unlink shm: %s", join.shm_name);
    }
 
-   // Parallelize.
-   Z_ (sem_post(join.sem));
+   Z_ (sem_post(join.sem));  // parallelize (all)
 
    Z_ (munmap(join.shared, sizeof(*join.shared)));
    Z_ (sem_close(join.sem));
@@ -339,12 +363,7 @@ void run_user_command(char *argv[], const char *initial_dir)
    if (initial_dir != NULL)
       Zf (chdir(initial_dir), "can't cd to %s", initial_dir);
 
-   if (verbose >= 3) {
-      fprintf(stderr, "argv:");
-      for (int i = 0; argv[i] != NULL; i++)
-         fprintf(stderr, " \"%s\"", argv[i]);
-      fprintf(stderr, "\n");
-   }
+   INFO("executing: %s", argv_to_string(argv));
 
    Zf (prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0), "can't set no_new_privs");
    execvp(argv[0], argv);  // only returns if error
@@ -427,7 +446,7 @@ void setup_passwd(const struct container *c)
    struct passwd *p;
 
    // /etc/passwd
-   T_ (path = strdup("/tmp/ch-run_passwd.XXXXXX"));
+   T_ (path = cat(host_tmp, "/ch-run_passwd.XXXXXX"));
    T_ (-1 != (fd = mkstemp(path)));  // mkstemp(3) writes path
    if (c->container_uid != 0)
       T_ (1 <= dprintf(fd, "root:x:0:0:root:/root:/bin/sh\n"));
@@ -438,7 +457,7 @@ void setup_passwd(const struct container *c)
    if (p) {
       T_ (1 <= dprintf(fd, "%s:x:%u:%u:%s:/home/%s:/bin/sh\n",
                        p->pw_name, c->container_uid, c->container_gid,
-                       p->pw_gecos, getenv("USER")));
+                       p->pw_gecos, username));
    } else {
       if (errno) {
          Tf (0, "getpwuid(3) failed");
@@ -454,7 +473,7 @@ void setup_passwd(const struct container *c)
    Z_ (unlink(path));
 
    // /etc/group
-   T_ (path = strdup("/tmp/ch-run_group.XXXXXX"));
+   T_ (path = cat(host_tmp, "/ch-run_group.XXXXXX"));
    T_ (-1 != (fd = mkstemp(path)));
    if (c->container_gid != 0)
       T_ (1 <= dprintf(fd, "root:x:0:\n"));
