@@ -20,12 +20,19 @@ struct bind {
    enum bind_dep dep;
 };
 
+enum img_type {
+   IMG_DIRECTORY,  // normal directory, perhaps an external mount of some kind
+   IMG_SQUASH,     // SquashFS archive file (not yet mounted)
+   IMG_NONE,       // image type is not set yet
+};
+
 struct container {
    struct bind *binds;
    bool ch_ssh;          // bind /usr/bin/ch-ssh?
    gid_t container_gid;  // GID to use in container
    uid_t container_uid;  // UID to use in container
-   bool env_expand;      // identifier to expands variable
+   bool env_expand;      // expand variables in --set-env
+   char *img_path;       // path to image
    char *newroot;        // path to new root directory
    bool join;            // is this a synchronized join?
    int join_ct;          // number of peers in a synchronized join
@@ -35,11 +42,13 @@ struct container {
    bool private_passwd;  // don't bind custom /etc/{passwd,group}
    bool private_tmp;     // don't bind host's /tmp
    char *old_home;       // host path to user's home directory (i.e. $HOME)
-   bool writable;
+   enum img_type type;   // directory, SquashFS, etc.
+   bool writable;        // re-mount image read-write
 };
 
 
 /** Function prototypes **/
 
 void containerize(struct container *c);
+enum img_type img_type_get(const char *path);
 void run_user_command(char *argv[], const char *initial_dir);
