@@ -57,7 +57,8 @@ ARG_DEFAULTS = { "HTTP_PROXY": os.environ.get("HTTP_PROXY"),
                  "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin",
                  # GNU tar, when it thinks it's running as root, tries to
                  # chown(2) and chgrp(2) files to whatever's in the tarball.
-                 "TAR_OPTIONS": "--no-same-owner" }
+                 "TAR_OPTIONS": "--no-same-owner",
+                 "USER": ch.user() }
 
 
 ## Main ##
@@ -131,7 +132,7 @@ def main(cli_):
    env = Environment()
 
    # Read input file.
-   if (cli.file == "-"):
+   if (cli.file == "-" or cli.context == "-"):
       text = ch.ossafe(sys.stdin.read, "can't read stdin")
    else:
       fp = ch.open_(cli.file, "rt")
@@ -504,6 +505,8 @@ class I_copy(Instruction):
       return dst_canon
 
    def execute_(self):
+      if (cli.context == "-"):
+         ch.FATAL("can't COPY: no context because \"-\" given")
       if (len(self.srcs) < 1):
          ch.FATAL("can't COPY: must specify at least one source")
       # Complain about unsupported stuff.
