@@ -191,10 +191,15 @@ HTTP_CHUNK_SIZE = 256 * 1024
 PYTHON_MIN = (3,6)
 
 # Content types for some stuff we care about.
-TYPE_MANIFEST = "application/vnd.docker.distribution.manifest.v2+json"
-TYPE_MANIFEST_LIST = "application/vnd.docker.distribution.manifest.list.v2+json"
-TYPE_CONFIG =   "application/vnd.docker.container.image.v1+json"
-TYPE_LAYER =    "application/vnd.docker.image.rootfs.diff.tar.gzip"
+# See: https://github.com/opencontainers/image-spec/blob/main/media-types.md
+TYPES_MANIFEST = \
+   {"docker2": "application/vnd.docker.distribution.manifest.v2+json",
+    "oci1":    "application/vnd.oci.image.manifest.v1+json"}
+TYPES_INDEX = \
+   {"docker2": "application/vnd.docker.distribution.manifest.list.v2+json",
+    "oci1":    "application/vnd.oci.image.index.v1+json"}
+TYPE_CONFIG = "application/vnd.docker.container.image.v1+json"
+TYPE_LAYER = "application/vnd.docker.image.rootfs.diff.tar.gzip"
 
 # Top-level directories we create if not present.
 STANDARD_DIRS = { "bin", "dev", "etc", "mnt", "proc", "sys", "tmp", "usr" }
@@ -254,7 +259,10 @@ class Credentials:
          except KeyError:
             # Finally, prompt the user.
             # FIXME: This hangs in Bats despite sys.stdin.isatty() == True.
-            username = input("\nUsername: ")
+            try:
+               username = input("\nUsername: ")
+            except KeyboardInterrupt:
+               FATAL("authentication cancelled")
             password = getpass.getpass("Password: ")
          if (not password_many):
             # Remember the credentials.
@@ -391,7 +399,7 @@ class Image:
    def metadata_init(self):
       "Initialize empty metadata structure."
       # Elsewhere can assume the existence and types of everything here.
-      self.metadata = { "arch": None,
+      self.metadata = { "arch": arch_host.split("/")[0],  # no variant
                         "cwd": "/",
                         "env": dict(),
                         "labels": dict(),
@@ -596,7 +604,7 @@ class Image:
          dev_ct = 0
          members2 = list(members)  # copy b/c we'll alter members
          for m in members2:
-            self.validate_tar_path(fp.name, m.name)
+            TarFile.fix_member_path(m, fp.name)
             if (m.isdev()):
                # Device or FIFO: Ignore.
                dev_ct += 1
@@ -623,17 +631,9 @@ class Image:
             INFO("layer %d/%d: %s: ignored %d devices and/or FIFOs"
                  % (i, len(layers), lh[:7], dev_ct))
 
-   def validate_tar_path(self, filename, path):
-      "Reject paths outside the tar top level by aborting the program."
-      if (len(path) > 0 and path[0] == "/"):
-         FATAL("rejecting absolute path: %s: %s" % (filename, path))
-      if (".." in path.split("/")):
-         FATAL("rejecting path with up-level: %s: %s" % (filename, path))
-
    def validate_tar_link(self, filename, path, target):
       """Reject hard link targets outside the tar top level by aborting the
          program."""
-      self.validate_tar_path(filename, path)
       if (len(target) > 0 and target[0] == "/"):
          FATAL("rejecting absolute hard link target: %s: %s -> %s"
                % (filename, path, target))
@@ -857,7 +857,15 @@ fields:
 
    def defaults_add(self):
       "Set defaults for all empty fields."
-      if (self.host is None): self.host = "registry-1.docker.io"
+      if (self.host is None):
+         if ("CH_REGY_DEFAULT_HOST" not in os.environ):
+            self.host = "registry-1.docker.io"
+         else:
+            self.host = os.getenv("CH_REGY_DEFAULT_HOST")
+            self.port = int(os.getenv("CH_REGY_DEFAULT_PORT", 443))
+            prefix = os.getenv("CH_REGY_PATH_PREFIX")
+            if (prefix is not None):
+               self.path = prefix.split("/") + self.path
       if (self.port is None): self.port = 443
       if (self.host == "registry-1.docker.io" and len(self.path) == 0):
          # FIXME: For Docker Hub only, images with no path need a path of
@@ -1322,11 +1330,12 @@ class Registry_HTTP:
          responsible for distinguishing cases 1 and 2."""
       url = self._url_of("manifests", self.ref.version)
       pw = Progress_Writer(path, msg)
-      # Including TYPE_MANIFEST avoids the server trying to convert its v2
+      # Including TYPES_MANIFEST avoids the server trying to convert its v2
       # manifest to a v1 manifest, which currently fails for images
       # Charliecloud pushes. The error in the test registry is “empty history
       # when trying to create schema1 manifest”.
-      accept = "%s, %s;q=0.5" % (TYPE_MANIFEST_LIST, TYPE_MANIFEST)
+      accept = "%s;q=0.5" % ",".join(  list(TYPES_INDEX.values())
+                                     + list(TYPES_MANIFEST.values()))
       res = self.request("GET", url, out=pw, statuses={200, 401, 404},
                          headers={ "Accept" : accept })
       pw.close()
@@ -1352,8 +1361,9 @@ class Registry_HTTP:
          digest = "sha256:" + digest
       url = self._url_of("manifests", digest)
       pw = Progress_Writer(path, msg)
+      accept = "%s;q=0.5" % ",".join(TYPES_MANIFEST.values())
       res = self.request("GET", url, out=pw, statuses={200, 401, 404},
-                         headers={ "Accept" : TYPE_MANIFEST })
+                         headers={ "Accept" : accept })
       pw.close()
       if (res.status_code != 200):
          DEBUG(res.content)
@@ -1365,7 +1375,7 @@ class Registry_HTTP:
       INFO("manifest: uploading")
       url = self._url_of("manifests", self.ref.tag)
       self.request("PUT", url, {201}, data=manifest,
-                   headers={ "Content-Type": TYPE_MANIFEST })
+                   headers={ "Content-Type": TYPES_MANIFEST["docker2"] })
 
    def request(self, method, url, statuses={200}, out=None, **kwargs):
       """Request url using method and return the response object. If statuses
@@ -1459,6 +1469,13 @@ class Storage:
    def upload_cache(self):
       return self.root // "ulcache"
 
+   @property
+   def valid_p(self):
+      # FIXME: require version file too when appropriate (#1147)
+      """Return True if storage present and seems valid, False otherwise."""
+      return (os.path.isdir(self.unpack_base) and
+              os.path.isdir(self.download_cache))
+
    @property
    def version_file(self):
       return self.root // "version"
@@ -1469,11 +1486,7 @@ class Storage:
       # environment variable? It seems a lot more robust. But, (1) we'd have
       # to match it in some scripts and (2) it makes the documentation less
       # clear becase we have to explain the fallback behavior.
-      try:
-         username = os.environ["USER"]
-      except KeyError:
-         FATAL("can't get username: $USER not set")
-      return "/var/tmp/%s/ch-image" % username
+      return Path("/var/tmp/%s.ch" % user())
 
    @staticmethod
    def root_env():
@@ -1488,6 +1501,7 @@ class Storage:
    def init(self):
       """Ensure the storage directory exists, contains all the appropriate
          top-level directories & metadata, and is the appropriate version."""
+      self.init_move_old()  # see issues #1160 and #1243
       if (not os.path.isdir(self.root)):
          op = "initializing"
          v_found = None
@@ -1503,7 +1517,7 @@ class Storage:
          VERBOSE("found storage dir v%d: %s" % (STORAGE_VERSION, self.root))
       elif (v_found in {None, 1}):  # initialize/upgrade
          INFO("%s storage directory: v%d %s" % (op, STORAGE_VERSION, self.root))
-         mkdirs(self.root)  # FIXME: use mkdir() when don't need two levels
+         mkdir(self.root)
          mkdir(self.download_cache)
          mkdir(self.unpack_base)
          mkdir(self.upload_cache)
@@ -1512,6 +1526,50 @@ class Storage:
          FATAL("incompatible storage directory v%d: %s" % (v_found, self.root),
                'you can delete and re-initialize with "ch-image reset"')
 
+   def init_move_old(self):
+      """If appropriate, move storage directory from old default path to new.
+         See issues #1160 and #1243."""
+      old = Storage("/var/tmp/%s/ch-image" % user())
+      moves = ( "dlcache", "img", "ulcache", "version" )
+      if (self.root != self.root_default()):
+         return  # do nothing silently unless using default storage dir
+      if (not os.path.exists(old.root)):
+         return  # do nothing silently unless old default storage dir exists
+      if (not old.valid_p):
+         WARNING("storage dir: invalid at old default, ignoring: %s" % old.root)
+         return
+      INFO("storage dir: valid at old default: %s" % old.root)
+      if (not os.path.exists(self.root)):
+         mkdir(self.root)
+      elif (self.valid_p):
+         WARNING("storage dir: also valid at new default: %s" % self.root,
+                 hint="consider deleting the old one")
+         return
+      elif (not os.path.isdir(self.root)):
+         return  # new isn't a directory; init/upgrade code will error later
+      elif (any(os.path.exists(self.root // i) for i in moves)):
+         return  # new is broken; init/upgrade should error later
+      # Now we know (1) the old storage exists and is valid and (2) the new
+      # storage exists, is a directory, and contains none of the files we'd
+      # move. However, it *may* contain subdirectories other parts of
+      # Charliecloud care about, e.g. "mnt" for ch-run.
+      INFO("storage dir: moving to new default path: %s" % self.root)
+      for i in moves:
+         src = old.root // i
+         dst = self.root // i
+         if (os.path.exists(src)):
+            DEBUG("moving: %s -> %s" % (src, dst))
+            try:
+               shutil.move(src, dst)
+            except OSError as x:
+               FATAL("can't move: %s -> %s: %s"
+                     % (x.filename, x.filename2, x.strerror))
+      ossafe(os.rmdir, "can't rmdir: %s" % old.root, old.root)
+      if (len(ossafe(os.listdir, "can't list: %s" % old.root.parent,
+                     old.root.parent)) == 0):
+         WARNING("parent of old storage dir now empty: %s" % old.root.parent,
+                 hint="consider deleting it")
+
    def manifest_for_download(self, image_ref, digest):
       if (digest is None):
          digest = "skinny"
@@ -1522,7 +1580,7 @@ class Storage:
       return self.download_cache // ("%s.fat.json" % image_ref.for_path)
 
    def reset(self):
-      if (self.valid_p()):
+      if (self.valid_p):
          rmtree(self.root)
          self.init()  # largely for debugging
       else:
@@ -1531,12 +1589,6 @@ class Storage:
    def unpack(self, image_ref):
       return self.unpack_base // image_ref.for_path
 
-   def valid_p(self):
-      # FIXME: require version file too when appropriate (#1147)
-      "Return True if storage present and seems valid, False otherwise."
-      return (os.path.isdir(self.unpack_base) and
-              os.path.isdir(self.download_cache))
-
 
 class TarFile(tarfile.TarFile):
 
@@ -1588,6 +1640,20 @@ class TarFile(tarfile.TarFile):
             FATAL("invalid file type 0%o in previous layer; see inode(7): %s"
                   % (stat.S_IFMT(st.st_mode), targetpath))
 
+   @staticmethod
+   def fix_member_path(ti, tarball_path):
+      """Repair or reject (by aborting the program) ti.name (member path) if
+         it climbs outside the top level or has some other problem."""
+      old_name = ti.name
+      if (len(ti.name) == 0):
+         FATAL("rejecting zero-length member path: %s" % (tarball_path))
+      if (".." in ti.name.split("/")):
+         FATAL("rejecting member path with up-level: %s: %s" % (filename, path))
+      if (ti.name[0] == "/"):
+         ti.name = "." + old_name  # add dot so we don't have to count slashes
+      if (ti.name != old_name):
+         WARNING("fixed member path: %s -> %s" % (old_name, ti.name))
+
    @staticmethod
    def fix_member_uidgid(ti):
       assert (ti.name[0] != "/")  # absolute paths unsafe but shouldn't happen
@@ -1731,7 +1797,7 @@ def digest_trim(d):
       FATAL("no algorithm tag: %s" % d)
 
 def done_notify():
-   if (os.environ.get("USER", None) == "jogas"):
+   if (user() == "jogas"):
       INFO("!!! KOBE !!!")
    else:
       INFO("done")
@@ -2011,3 +2077,10 @@ def tree_terminals_cat(tree, tname):
 def unlink(path, *args, **kwargs):
    "Error-checking wrapper for os.unlink()."
    ossafe(os.unlink, "can't unlink: %s" % path, path)
+
+def user():
+   "Return the current username; exit with error if it can't be obtained."
+   try:
+      return os.environ["USER"]
+   except KeyError:
+      FATAL("can't get username: $USER not set")
